from trace_dump cimport (TraceDump, FileTraceDump, SyslogTraceDump, SyslogSocket, SyslogUNIXSocket, SyslogTCPSocket)
from libc.stdio cimport FILE, fopen, fclose, stdout, stderr

cdef class PyTraceDump:
    cdef TraceDump* thisptr

    def __dealloc__(self):
        del self.thisptr

    def start(self):
        self.thisptr.start()

    def stop(self):
        self.thisptr.stop()


cdef class PyFileTraceDump(PyTraceDump):
    def __cinit__(self, const char* path):
        global trace_message_ring_buffer
        cdef FILE* handle = fopen(path, "wb")
        if handle == NULL:
            raise ValueError("cannot open file {}".format(path))
        self.thisptr = new FileTraceDump(trace_message_ring_buffer, handle, True)


cdef class PyStdoutTraceDump(PyTraceDump):
    def __cinit__(self):
        global trace_message_ring_buffer
        self.thisptr = new FileTraceDump(trace_message_ring_buffer, stdout, False)


cdef class PyStderrTraceDump(PyTraceDump):
    def __cinit__(self):
        global trace_message_ring_buffer
        self.thisptr = new FileTraceDump(trace_message_ring_buffer, stderr, False)


cdef class PySyslogTraceDump(PyTraceDump):
    def __cinit__(self):
        self.thisptr = NULL


cdef new_syslog_trace_dump(const char* host_name, const char* application_name, const char* process_id, int facility,
                           bool rfc5424, SyslogSocket* socket):
    global trace_message_ring_buffer
    cdef SyslogTraceDump* trace_dump = new SyslogTraceDump(trace_message_ring_buffer, host_name, application_name,
                                                           process_id, facility, rfc5424, socket)
    result = PySyslogTraceDump()
    result.thisptr = trace_dump
    return result


def PySyslogTraceDump_create_with_unix_socket(const char* host_name, const char* application_name,
                                              const char* process_id, int facility, bool rfc5424, const char* address):
    cdef SyslogUNIXSocket* socket = new SyslogUNIXSocket(address)
    return new_syslog_trace_dump(host_name, application_name, process_id, facility, rfc5424, socket)


def PySyslogTraceDump_create_with_tcp_socket(const char* host_name, const char* application_name,
                                             const char* process_id, int facility, bool rfc5424, const char* address,
                                             int port):
    cdef SyslogTCPSocket* socket = new SyslogTCPSocket(address, port)
    return new_syslog_trace_dump(host_name, application_name, process_id, facility, rfc5424, socket)


cdef class PyWriter:
    cdef TraceMessageRingBuffer* ring_buffer
    cdef TraceDump* trace_dump

    def __cinit__(self):
        self.ring_buffer = NULL
        self.trace_dump = NULL

    def __dealloc__(self):
        self.trace_dump.stop()
        del self.trace_dump
        del self.ring_buffer

    def start(self):
        self.trace_dump.start()

    def stop(self):
        self.trace_dump.stop()

    def write(self, severity, message_str):
        cdef TraceMessage* message = self.ring_buffer.reserve_push()
        message.set_timestamp()
        message.set_severity(severity)
        message.write(message_str)
        self.ring_buffer.commit_push(message)


cdef new_syslog_writer(int buffer_size, int trace_message_size, const char* host_name, const char* application_name, 
                       const char* process_id, int facility, bool rfc5424, SyslogSocket* socket):
    cdef TraceMessageRingBuffer* ring_buffer = new TraceMessageRingBuffer(buffer_size, trace_message_size)
    cdef SyslogTraceDump* trace_dump = new SyslogTraceDump(ring_buffer, host_name, application_name, process_id,
                                                           facility, rfc5424, socket)
    result = PyWriter()
    result.ring_buffer = ring_buffer
    result.trace_dump = trace_dump
    return result

def PySyslogWriter_create_with_unix_socket(int buffer_size, int trace_message_size, const char* host_name, 
                                           const char* application_name, const char* process_id, int facility, 
                                           bool rfc5424, const char* address):
    cdef SyslogSocket* socket = new SyslogUNIXSocket(address)
    return new_syslog_writer(buffer_size, trace_message_size, host_name, application_name, process_id, facility, 
                             rfc5424, socket)


def PySyslogWriter_create_with_tcp_socket(int buffer_size, int trace_message_size, const char* host_name, 
                                          const char* application_name, const char* process_id, int facility, 
                                          bool rfc5424, const char* address, int port):
    cdef SyslogSocket* socket = new SyslogTCPSocket(address, port)
    return new_syslog_writer(buffer_size, trace_message_size, host_name, application_name, process_id, facility, 
                             rfc5424, socket)



cdef extern from "lru.hpp" namespace "plb":
    cdef cppclass LRUCacheH4ConstIterator[K, V]:
        const K& key() nogil
        const V& value() nogil
        bool operator==(const LRUCacheH4ConstIterator[K, V]& other) nogil
        bool operator!=(const LRUCacheH4ConstIterator[K, V]& other) nogil

    cdef cppclass LRUCacheH4[K, V]:
        LRUCacheH4(int)
        LRUCacheH4ConstIterator[K, V] find(const K& key) nogil
        LRUCacheH4ConstIterator[K, V] end() nogil
        void insert(const K& key, const V& value) nogil


ctypedef LRUCacheH4[long,int] CodeLRUCache
ctypedef LRUCacheH4ConstIterator[long,int] CodeLRUCacheConstIterator

cdef CodeLRUCache* trace_level_func_cache = NULL
cdef unsigned long func_cache_hit, func_cache_miss

cdef inline int call_filter_and_store_trace_level(PyObject* filter_func, PyFrameObject* frame) with gil:
    global trace_level_func_cache
    cdef int trace_level

    trace_level_result = (<object>filter_func)(<object>frame)
    if trace_level_result is None:
        # TODO this is an error and we should log it.
        import sys
        sys.stderr.write("infi.tracing: filter returned None for frame: {} (filter: {})".format(<object>frame.f_code, 
                                                                                                <object>filter_func))
        trace_level_result = NO_TRACE
    trace_level = int(trace_level_result)
    trace_level_func_cache.insert(<long>frame.f_code, trace_level_result)
    return trace_level


cdef inline int find_trace_level_or_call_filter(PyFrameObject* frame, PyObject* filter_func) nogil:
    global trace_level_func_cache
    cdef int trace_level
    cdef CodeLRUCacheConstIterator code_found

    code_found = trace_level_func_cache.find(<long>frame.f_code)
    if code_found != trace_level_func_cache.end():
        trace_level = code_found.value()
        inc(func_cache_hit)
    else:
        trace_level = call_filter_and_store_trace_level(filter_func, frame)
        inc(func_cache_miss)
    return trace_level



__version__ = "0.2"
__git_commiter_name__ = "Tal Yalon"
__git_commiter_email__ = "yalon@infinidat.com"
__git_branch__ = '(detached from v0.2)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '34ecabca48b79636bbbca4350aaea1ffe22f33cc'
__git_head_subject__ = 'removed dependency on boost'
__git_head_message__ = ''
__git_dirty_diff__ = ''


from cython.operator cimport dereference as deref, preincrement as inc, predecrement as dec

from defs cimport *
from thread_storage cimport (ThreadStorage, GreenletStorage, NO_TRACE_FROM_DEPTH_DISABLED,
                             init_thread_storage_once, get_thread_storage)
from trace_message_ring_buffer cimport TraceMessageRingBuffer
from wait_and_ensure_exit cimport WaitAndEnsureExit

from os import getpid

DEFAULT_FUNC_CACHE_SIZE = 10000

cdef enum:
    NO_TRACE              = 0
    NO_TRACE_NESTED       = 1
    TRACE_FUNC_NAME       = 2
    TRACE_FUNC_PRIMITIVES = 3
    TRACE_FUNC_REPR       = 4


cdef unsigned long pid = -1
cdef unsigned long gid_hit = 0, gid_miss = 0
cdef unsigned long gstore_hit = 0, gstore_miss = 0

# For traces we have a 32k records buffer, each record is 2k long = 64MB buffer
cdef TraceMessageRingBuffer* trace_message_ring_buffer = new TraceMessageRingBuffer(32768, 2048)

include "trace_level_func_cache.pyx"
include "ctracing_log.pyx"
include "ctracing_trace_dump.pyx"

cdef inline long calc_new_greenlet_depth(PyFrameObject* frame) nogil:
    cdef long depth = 0
    while frame.f_back != NULL:
        inc(depth)
        frame = frame.f_back
    return depth

# Apparently when adding a GIL context to a function, Cython creates on the function's return that adds to the runtime
# cost even if the GIL context wasn't reached inside the function.
# That's why we had to create this function, so GIL won't get called if this isn't called.
cdef inline long get_current_gid() with gil:
    cdef PyGreenlet* g = PyGreenlet_GetCurrent()
    Py_DECREF(<PyObject*> g)
    return <long> g

cdef inline GreenletStorage* get_gstore_on_call(ThreadStorage* tstore, PyFrameObject* frame) nogil:
    global gid_hit, gid_miss, gstore_hit, gstore_miss
    cdef:
        long gid
        GreenletStorage* gstore

    # Easy route: if we're in the same greenlet we may have the greenlet ID cached.
    if (tstore.last_frame != 0) and (tstore.last_frame == <long>frame.f_back):
        gid = tstore.last_gid
        gstore = tstore.last_gstorage
        inc(gid_hit)
    else:
        inc(gid_miss)
        gid = get_current_gid()
        gstore = NULL

    tstore.last_gid = gid
    tstore.last_frame = <long>frame

    if gstore == NULL:
        inc(gstore_miss)
        gstore = tstore.find_gstorage(gid)
        if gstore == NULL:
            gstore = tstore.new_gstorage(gid)
            gstore.depth = calc_new_greenlet_depth(frame)
        tstore.last_gstorage = gstore
    else:
        inc(gstore_hit)

    return gstore


cdef inline GreenletStorage* get_gstore_on_return(ThreadStorage* tstore, PyFrameObject* frame) nogil:
    global gid_hit, gid_miss, gstore_hit, gstore_miss
    cdef:
        long gid
        GreenletStorage* gstore

    # Easy route: if we're in the same greenlet we may have the greenlet ID cached.
    if (tstore.last_frame != 0) and (tstore.last_frame == <long>frame):
        gid = tstore.last_gid
        gstore = tstore.last_gstorage
        inc(gid_hit)
    else:
        inc(gid_miss)
        gid = get_current_gid()
        gstore = NULL

    tstore.last_gid = gid
    tstore.last_frame = <long>frame.f_back

    if gstore == NULL:
        inc(gstore_miss)
        gstore = tstore.find_gstorage(gid)
        if gstore == NULL:
            gstore = tstore.new_gstorage(gid)
            gstore.depth = calc_new_greenlet_depth(frame)
        tstore.last_gstorage = gstore
    else:
        inc(gstore_hit)

    return gstore


cdef int greenlet_trace_func(PyObject* filter_func, PyFrameObject* frame, int what, PyObject* arg) nogil:
    cdef:
        int trace_level = NO_TRACE
        unsigned long tid
        long gid, depth, no_trace_from_depth
        ThreadStorage* tstore
        GreenletStorage* gstore = NULL

    tstore = get_thread_storage()
    tid = tstore.id

    if what == PyTrace_CALL:
        gstore = get_gstore_on_call(tstore, frame)
        inc(gstore.depth)
        depth = gstore.depth
        no_trace_from_depth = gstore.no_trace_from_depth
        gid = gstore.gid
    elif what == PyTrace_RETURN:
        gstore = get_gstore_on_return(tstore, frame)
        depth = gstore.depth
        no_trace_from_depth = gstore.no_trace_from_depth
        gid = gstore.gid

        if frame.f_back == NULL:
            tstore.del_gstorage(gstore)
            gstore = NULL
        else:
            dec(gstore.depth)
    else:
        # We don't trace through C calls, so we don't check the following:
        # PyTrace_C_CALL, PyTrace_C_EXCEPTION, PyTrace_C_RETURN
        # Also, PyTrace_EXCEPTION cannot happen with a profile function.
        return 0

    if tstore.enabled <= 0:
        return 0

    if no_trace_from_depth != NO_TRACE_FROM_DEPTH_DISABLED:
        if no_trace_from_depth < depth:
            return 0
        elif gstore != NULL:
            gstore.no_trace_from_depth = NO_TRACE_FROM_DEPTH_DISABLED

    trace_level = find_trace_level_or_call_filter(frame, filter_func)

    if trace_level == NO_TRACE:
        return 0
    elif trace_level == NO_TRACE_NESTED:
        if gstore != NULL:
            gstore.no_trace_from_depth = depth
        return 0

    if what == PyTrace_CALL:
        log_call(trace_level, tid, gid, depth, frame, arg)
    else:
        log_return(trace_level, tid, gid, depth, frame, arg)

    return 0


def ctracing_set_profile(filter_func):
    global pid, trace_level_func_cache, _PyGreenlet_API

    init_thread_storage_once()
    pid = <long> getpid()  # We set this once per process so we won't call this method on every trace.

    cdef ThreadStorage* tstore

    if filter_func is None:
        raise ValueError("filter_func cannot be None")

    PyGreenlet_Import()
    if trace_level_func_cache == NULL:
        ctracing_set_func_cache_size(DEFAULT_FUNC_CACHE_SIZE)

    if _PyGreenlet_API != NULL:
        PyEval_SetProfile(greenlet_trace_func, <PyObject*>filter_func)
    else:
        raise ValueError("tracing only supported with greenlets for now")


def ctracing_set_func_cache_size(size):
    global trace_level_func_cache
    assert size >= 0

    if trace_level_func_cache != NULL:
        del trace_level_func_cache

    trace_level_func_cache = new CodeLRUCache(size)
    if trace_level_func_cache == NULL:
        raise Exception("failed to create trace level func cache (size: {})".format(size))


def ctracing_print_stats():
    global gid_hit, gid_miss, gstore_hit, gstore_miss, func_cache_hit, func_cache_miss
    print("gid hits: {}, misses: {}".format(gid_hit, gid_miss))
    print("gstore hits: {}, misses: {}".format(gstore_hit, gstore_miss))
    print("func cache hits: {}, misses: {}".format(func_cache_hit, func_cache_miss))
    print("overflow message counter: {}".format(trace_message_ring_buffer.get_overflow_counter()))
    print("spinlock counters: consumer {} / producer {}".format(
        trace_message_ring_buffer.get_spinlock_consumer_wait_counter(),
        trace_message_ring_buffer.get_spinlock_producer_wait_counter()))


def suspend():
    cdef ThreadStorage* tstore = get_thread_storage()
    dec(tstore.enabled)


def resume():
    cdef ThreadStorage* tstore = get_thread_storage()
    if tstore.enabled <= 0:
        inc(tstore.enabled)


def ctracing_wait_and_ensure_exit(seconds, exit_code):
    cdef WaitAndEnsureExit* obj = new WaitAndEnsureExit()
    obj.go(seconds, exit_code)



from defs cimport *
from trace_message cimport TraceMessage

include "ctracing_log_serialize.pyx"

cdef char* UNKNOWN_MODULE = "<unknown>"

cdef inline bool serialize_prefix(char direction, long tid, long gid, long depth, PyFrameObject* frame, 
                                  TraceMessage* message) with gil:
    global pid
    cdef:
        char* file_name
        char* func_name
        char* module_name

    line_no = PyFrame_GetLineNumber(frame)
    file_name = PyString_AsString(frame.f_code.co_filename)
    func_name = PyString_AsString(frame.f_code.co_name)

    module_name_obj = (<object> frame.f_globals).get('__name__')
    if module_name_obj is not None:
        module_name = PyString_AsString(<PyObject*>module_name_obj)
    else:
        module_name = UNKNOWN_MODULE

    return message.printf("%c,%d:%lu:%d:%d,%s:%d,%s,%s", direction, pid, tid % 32768, gid % 32768, depth, 
                          file_name, line_no, module_name, func_name)


cdef void log_call(int trace_level, long tid, long gid, long depth, PyFrameObject* frame, PyObject* arg) nogil:
    global trace_message_ring_buffer
    cdef:
        int locals_i = 0
        TraceMessage* trace_message = trace_message_ring_buffer.reserve_push()

    if trace_message == NULL:
        return

    try:
        if not serialize_prefix('>', tid, gid, depth, frame, trace_message):
            # TODO: not enough room to write this down, maybe we should write TRUNCATED or something like that.
            return

        if trace_level > TRACE_FUNC_NAME:
            with gil:
                for locals_i in range(frame.f_code.co_argcount):
                    trace_message.write(",")
                    fast_repr(frame.f_localsplus[locals_i], trace_message)

                locals_i = frame.f_code.co_argcount
                if frame.f_code.co_flags & CO_VARARGS:
                    trace_message.write(",vargs=")
                    fast_repr(frame.f_localsplus[locals_i], trace_message)
                    inc(locals_i)

                if frame.f_code.co_flags & CO_VARKEYWORDS:
                    trace_message.write(",kwargs=")
                    fast_repr(frame.f_localsplus[locals_i], trace_message)
                    inc(locals_i)
    finally:
        trace_message.set_timestamp()
        trace_message_ring_buffer.commit_push(trace_message)

cdef void log_return(int trace_level, long tid, long gid, long depth, PyFrameObject* frame, PyObject* arg) nogil:
    global trace_message_ring_buffer
    cdef TraceMessage* trace_message = trace_message_ring_buffer.reserve_push()
    cdef PyObject* exc_type = frame.f_tstate.exc_type
    cdef PyObject* exc_value = frame.f_tstate.exc_value

    if trace_message == NULL:
        return        

    try:
        if not serialize_prefix('<', tid, gid, depth, frame, trace_message):
            # TODO: not enough room to write this down, maybe we should write TRUNCATED or something like that.
            return

        if trace_level > TRACE_FUNC_NAME:
            trace_message.write(",")
            with gil:
                if arg != NULL:
                    fast_repr(arg, trace_message)
                else:
                    # Since we're using setprofile, we can't know what exception was raised because Python's code
                    # (specifically ceval.c:call_trace_protected) masks the exception before calling us - it calls
                    # PyErr_Fetch that clears the exception from the frame.
                    trace_message.write("exc")
    finally:
        trace_message.set_timestamp()
        trace_message_ring_buffer.commit_push(trace_message)



from libcpp cimport bool

cdef extern from "thread_storage.h":
    cdef enum: NO_TRACE_FROM_DEPTH_DISABLED "NO_TRACE_FROM_DEPTH_DISABLED"

    cdef cppclass GreenletStorage:
        long gid
        long depth
        long no_trace_from_depth
        bool enabled

    cdef cppclass ThreadStorage:
        unsigned long id
        int enabled
        long last_frame
        long last_gid
        GreenletStorage* last_gstorage

        GreenletStorage* find_gstorage(long gid) nogil
        GreenletStorage* new_gstorage(long gid) nogil
        void del_gstorage(GreenletStorage* ptr) nogil

    ThreadStorage* get_thread_storage() nogil
    void init_thread_storage_once() nogil



from setuptools.extension import Extension

import sys
if 'setuptools.extension' in sys.modules:
    m = sys.modules['setuptools.extension']
    m.Extension.__dict__ = m._Extension.__dict__


SETUP_INFO = dict(
    name = 'infi.tracing',
    version = '0.2',
    author = 'Tal Yalon',
    author_email = 'yalon@infinidat.com',

    url = 'https://github.com/Infinidat/infi.tracing',
    license = 'PSF',
    description = """short description here""",
    long_description = """long description here""",

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Python Software Foundation License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    install_requires = ['cython',
'greenlet',
'infi.pyutils',
'setuptools'],
    setup_requires = ['setuptools_cython'],
    namespace_packages = ['infi'],

    package_dir = {'': 'src'},
    package_data = {'': ['*.pxd', '*.h', 'greenlet.h', '*.hpp', '*.pyx']},
    include_package_data = True,
    zip_safe = False,

    entry_points = dict(
        console_scripts = [],
        gui_scripts = [],
        ),

)

if SETUP_INFO['url'] is None:
    _ = SETUP_INFO.pop('url')


def build_ext_modules():
    from platform import platform, system
    extra_compile_args = ["-Wno-format-security"]
    extra_link_args = []
    if platform().startswith("Darwin-13"):
        extra_compile_args += ["-stdlib=libstdc++"]
        extra_link_args += ["-stdlib=libstdc++"]
    if system() in ["Linux", "Darwin"]:
        extra_compile_args += ["-Wno-strict-prototypes"]
        platform_specific_sources = ["src/mintsystem/gcc/datetime.c", "src/mintsystem/gcc/mutex.c",
                                     "src/mintsystem/gcc/semaphore.c", "src/mintsystem/gcc/timer.c"]
    else:
        platform_specific_sources = ["src/mintsystem/msvc/datetime.c", "src/mintsystem/msvc/thread.c",
                                     "src/mintsystem/msvc/timer.c"]
    return [Extension("infi.tracing.ctracing",
               language="c++",
               sources=["src/infi/tracing/ctracing.pyx", "src/infi/tracing/thread_storage.cpp",
                        "src/infi/tracing/trace_dump.cpp", "src/infi/tracing/wait_and_ensure_exit.cpp",
                        "src/mintomic/mintomic_gcc.c"] + platform_specific_sources,
               include_dirs=["src/infi/tracing", "src"],
               define_macros=[("_REENTRANT", 1)],
               libraries=[],
               extra_compile_args=extra_compile_args,
               extra_link_args=extra_link_args)]


def setup():
    from setuptools import setup as _setup
    from setuptools import find_packages
    SETUP_INFO['ext_modules'] = build_ext_modules()
    SETUP_INFO['packages'] = find_packages('src')
    _setup(**SETUP_INFO)

if __name__ == '__main__':
    setup()
